import {
  RegionType,
  ApiCredentials,
  removeUndefined,
  deepClone,
  ApiTypeError,
} from '@tencent-serverless-framework/common';
import { Client } from 'tencentcloud-sdk-nodejs/tencentcloud/services/vpc/v20170312/vpc_client';
import {
  CreateSubnetRequest,
  CreateVpcRequest,
  ModifySubnetAttributeRequest,
  ModifyVpcAttributeRequest,
} from 'tencentcloud-sdk-nodejs/tencentcloud/services/vpc/v20170312/vpc_models';
import { VpcDeployInputs, VpcRemoveInputs, VpcDeployOutputs } from './interface';

/** 腾讯云私有网络 （Vpc）for serverless component */
export class Vpc {
  /** 地域 */
  region: RegionType;
  /** 鉴权信息 */
  credentials: ApiCredentials;
  /** sdk 实例 */
  client: Client;

  /**
   * 构造函数
   *
   * @param credentials 认证信息
   * @param region 地域信息
   */
  constructor(credentials: ApiCredentials = {}, region: RegionType = 'ap-guangzhou') {
    this.region = region;
    this.credentials = credentials;

    this.client = new Client({
      credential: removeUndefined({
        secretId: credentials.SecretId,
        secretKey: credentials.SecretKey,
        token: credentials.Token,
      }),
      region: region,
      profile: {},
    });
  }

  /**
   * 部署私有网络
   *
   * @param inputs 输入
   * @returns 输出
   */
  async deploy(inputs: VpcDeployInputs): Promise<VpcDeployOutputs> {
    const {
      zone,
      vpcName,
      subnetName,
      cidrBlock,
      enableMulticast,
      dnsServers,
      domainName,
      tags,
      subnetTags,
      enableSubnetBroadcast,
    } = inputs;

    let { vpcId, subnetId } = inputs;

    const deployVpc = async (vId: string) => {
      let existVpc = false;
      if (vId) {
        const detail = await this.getVpcDetail(vId);
        if (detail) {
          existVpc = true;
        }
      }

      /** 修改旧的 Vpc */
      if (existVpc) {
        console.log(`Updating vpc ${vId}...`);
        await this.modifyVpc({
          VpcName: vpcName,
          EnableMulticast: enableMulticast,
          DnsServers: dnsServers,
          DomainName: domainName,
          VpcId: vId,
        });
        console.log(`Update vpc ${vId} success`);
      } else {
        /** 添加新的 Vpc */
        if (!cidrBlock) {
          throw new ApiTypeError('PARAMETER_VPC', 'cidrBlock is required');
        }

        console.log(`Creating vpc ${vpcName}...`);
        const res = await this.createVpc({
          VpcName: vpcName,
          EnableMulticast: enableMulticast,
          DnsServers: dnsServers,
          DomainName: domainName,
          CidrBlock: cidrBlock,
          Tags: tags,
        });
        console.log(`Create vpc ${vpcName} success.`);
        vId = res?.VpcId!;
      }
      return vId;
    };

    // check subnetId
    const deploySubnet = async (vId: string, sId: string) => {
      let existSubnet = false;
      if (sId) {
        const detail = await this.getSubnetDetail(sId);
        if (detail) {
          existSubnet = true;
        }
      }

      /** 子网已存在 */
      if (existSubnet) {
        console.log(`Updating subnet ${sId}...`);

        await this.modifySubnet({
          SubnetId: sId,
          EnableBroadcast: enableSubnetBroadcast,
          SubnetName: subnetName,
        });
        console.log(`Update subnet ${sId} success.`);
      } else {
        /** 子网不存在 */
        if (vId) {
          const res = await this.createSubnet(
            removeUndefined({
              SubnetName: subnetName,
              Zone: zone,
              VpcId: vId,
              CidrBlock: cidrBlock!,
              Tags: subnetTags,
            }),
          );
          sId = res?.SubnetId!;

          // work for string and boolean
          if ('' + enableSubnetBroadcast === 'true') {
            await this.modifySubnet({
              SubnetId: sId,
              EnableBroadcast: enableSubnetBroadcast,
            });
          }
          console.log(`Create subnet ${subnetName} success.`);
        }
      }
      return sId;
    };

    if (vpcName) {
      vpcId = await deployVpc(vpcId!);
    }

    if (subnetName) {
      subnetId = await deploySubnet(vpcId!, subnetId!);
    }

    return {
      region: this.region,
      zone,
      vpcId,
      vpcName,
      subnetId,
      subnetName,
    };
  }

  /**
   * 移除私有网络
   *
   * @param inputs 输入
   * @returns 成功时返回 {}
   */
  async remove(inputs: VpcRemoveInputs) {
    const { vpcId, subnetId } = inputs;
    if (subnetId) {
      console.log(`Start removing subnet ${subnetId}`);
      try {
        await this.deleteSubnet(subnetId);
      } catch (e) {
        console.warn(e);
      }
      console.log(`Removed subnet ${subnetId}`);
    }
    if (vpcId) {
      console.log(`Start removing vpc ${vpcId}`);
      try {
        await this.deleteVpc(vpcId);
      } catch (e) {
        console.warn(e);
      }
      console.log(`Removed vpc ${vpcId}`);
    }

    return {};
  }

  /**
   * 获取 VPC 详情
   *
   * @param vpcId VPC ID
   * @returns VPC 详情
   */
  async getVpcDetail(vpcId: string) {
    // get instance detail
    try {
      const res = await this.client.DescribeVpcs({
        VpcIds: [vpcId],
      });
      if (res.VpcSet) {
        const {
          VpcSet: [detail],
        } = res;
        return detail;
      }
      return null;
    } catch (e) {
      console.warn(e);
      return null;
    }
  }

  /**
   * 获取子网详情
   *
   * @param subnetId 子网 Id
   * @returns 子网详情
   */
  async getSubnetDetail(subnetId: string) {
    try {
      const res = await this.client.DescribeSubnets({
        SubnetIds: [subnetId],
      });
      if (res.SubnetSet) {
        const {
          SubnetSet: [detail],
        } = res;
        return detail;
      }
      return null;
    } catch (e) {
      console.warn(e);
      return null;
    }
  }

  /**
   * 创建 Vpc
   *
   * @param req 请求体
   * @returns 创建的 VPC
   */
  async createVpc(req: CreateVpcRequest) {
    // clean undefined
    req = deepClone(req);
    const res = await this.client.CreateVpc(req);
    if (res.Vpc && res.Vpc.VpcId) {
      return res.Vpc;
    }
  }

  /**
   * 修改 Vpc
   *
   * @param req 请求体
   * @returns 创建的 VPC
   */
  async modifyVpc(req: ModifyVpcAttributeRequest) {
    // clean undefined
    req = deepClone(req);
    await this.client.ModifyVpcAttribute(req);
  }

  /**
   * 删除 Vpc
   *
   * @param vpcId VPC ID
   */
  async deleteVpc(vpcId: string) {
    await this.client.DeleteVpc({
      VpcId: vpcId,
    });
  }

  /**
   * 创建子网
   *
   * @param req 请求体
   * @returns 创建的子网
   */
  async createSubnet(req: CreateSubnetRequest) {
    // clean undefined
    req = deepClone(req);
    const res = await this.client.CreateSubnet(req);
    if (res.Subnet && res.Subnet.SubnetId) {
      const { Subnet } = res;
      return Subnet;
    }
  }

  /**
   * 修改子网
   *
   * @param req 请求体
   */
  async modifySubnet(req: ModifySubnetAttributeRequest) {
    // clean undefined
    req = deepClone(req);
    await this.client.ModifySubnetAttribute(req);
  }

  /**
   * 删除子网
   *
   * @param subnetId 子网 ID
   */
  async deleteSubnet(subnetId: string) {
    await this.client.DeleteSubnet({
      SubnetId: subnetId,
    });
  }
}
